/*
 * Copyright (c) 1991-2024 by STEP Tools Inc. 
 * All Rights Reserved.
 * 
 * Permission to use, copy, modify, and distribute this software and
 * its documentation is hereby granted, provided that this copyright
 * notice and license appear on all copies of the software.
 * 
 * STEP TOOLS MAKES NO REPRESENTATIONS OR WARRANTIES ABOUT THE
 * SUITABILITY OF THE SOFTWARE, EITHER EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, OR NON-INFRINGEMENT. STEP TOOLS
 * SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED BY LICENSEE AS A
 * RESULT OF USING, MODIFYING OR DISTRIBUTING THIS SOFTWARE OR ITS
 * DERIVATIVES.
 *
 * Author: Dave Loffredo (loffredo@steptools.com)
 */

#include <rose.h>
#include <stp_schema.h>


void find_owning_representations ( 
    stp_representation_item * it,
    rose_vector &reps
    );

stp_product_definition_shape * find_owning_or_related_pdef_shape ( 
    stp_representation * rep
    );

stp_product_definition_shape * find_owning_pdef_shape ( 
    stp_representation * rep
    );



//============================================================
// FIND REPRESENTATION ITEMS -- After reading in a geometry source
// file, we use the following to find some faces on it for us to apply
// datums and tolerances to.

stp_advanced_face * find_face_by_name (
    RoseDesign * d, const char * it_name
    )
{
    // This is not the most efficient, but if ever used heavily it
    // could be improved by indexing the names with a hash table.

    RoseCursor cursor;
    RoseObject * obj;
    cursor.traverse (d);
    cursor.domain (ROSE_DOMAIN(stp_advanced_face));
    while ((obj = cursor.next()) != 0) {
	stp_advanced_face * it = ROSE_CAST(stp_advanced_face,obj);
	if (it-> name() && *(it-> name()) &&
	    !strcmp (it_name, it->name()))
	    return it;
    }
    return 0;
}


stp_advanced_face * find_face_by_p21id (
    RoseDesign * d, unsigned long id
    )
{
    RoseObject * obj = d-> findByEntityId(id);

    // This is null safe, and returns null if not a face
    return ROSE_CAST(stp_advanced_face,obj);
}





//============================================================
// FIND PRODUCT DEFINITION SHAPE -- Find the product definition shape
// property associated with a given representation item.  We use this
// to attach shape aspects.
// 
// This takes a little bit of work, because first we have to find the
// representation that uses the item, and that can require tracing up
// through several levels of nested rep items.
//
// Once we get the representation, we can look for a shape property
// that attaches the representation to a product definition, but we
// also need to watch for representations that are related to such a
// representation.
//
stp_product_definition_shape * find_pdef_shape_for_repitem ( 
    stp_representation_item * it
    )
{
    unsigned i,sz;
    rose_vector reps;
    stp_product_definition_shape * pds = 0;

    ROSE.beginTraversal();
    find_owning_representations (it, reps);

    // look for any product def that uses the rep directly as the
    // representation of a shape property, or one that uses a rep
    // related to one of these by a representation relationship
    //
    for (i=0, sz=reps.size(); i<sz; i++) {
	stp_representation * rep = (stp_representation*) reps[i];
	if ((pds = find_owning_or_related_pdef_shape(rep)) != 0)
	    break;
    }
    ROSE.endTraversal();

    return pds;
}


void find_owning_representations ( 
    stp_representation_item * it,
    rose_vector &reps
    )
{
    unsigned i,sz;
    RoseBackptrs *bp = rose_get_backptrs(it);
    for (i=0, sz=bp->size(); i<sz; i++) {
	RoseObject * ref = bp->get(i);
	if (ref-> wasVisited()) continue;

	if (ref-> isa(ROSE_DOMAIN(stp_representation))) {
	    ref-> visit();
	    reps.append(ROSE_CAST(stp_representation,ref));
	}
	else if (ref-> isa(ROSE_DOMAIN(stp_representation_item))) {
	    ref-> visit();
	    find_owning_representations (
		ROSE_CAST(stp_representation_item,ref), reps
		);
	}
    }
}


stp_representation * get_reprel_1(
    stp_representation_relationship * rel
    )
{
    if (!rel) return 0;
    RoseObject * obj = rose_get_nested_object (
	rel-> rep_1(), ROSE_DOMAIN(stp_representation)
	);
    return ROSE_CAST(stp_representation,obj);
}



stp_representation * get_reprel_2(
    stp_representation_relationship * rel
    )
{
    if (!rel) return 0;
    RoseObject * obj = rose_get_nested_object (
	rel-> rep_2(), ROSE_DOMAIN(stp_representation)
	);
    return ROSE_CAST(stp_representation,obj);
}
  

stp_product_definition_shape * find_owning_or_related_pdef_shape ( 
    stp_representation * rep
    )
{
    unsigned i,sz;

    stp_product_definition_shape * pds = find_owning_pdef_shape(rep);
    if (pds) return pds;

    // Look at representations that are related to this one and see if
    // they are directly owned by a product definition.

    RoseBackptrs *bp = rose_get_backptrs(rep);
    for (i=0, sz=bp->size(); i<sz; i++) {
	RoseObject * ref = bp->get(i);
	if (ref-> isa(ROSE_DOMAIN(stp_shape_representation_relationship))) {
	    stp_shape_representation_relationship * rr = 
		ROSE_CAST(stp_shape_representation_relationship,ref);

	    // Our representation could be on either side of the
	    // relationship.  Recursing could create an infinite loop
	    // if we did not mark the representations we visit.
	    stp_representation * related_rep = 
		((get_reprel_2(rr) == rep)? get_reprel_1(rr): get_reprel_2(rr));

	    if (related_rep-> wasVisited()) continue;
	    related_rep-> visit();

	    pds = find_owning_or_related_pdef_shape (related_rep);
	    if (pds) return pds;
	}
    }
    return 0;
}
  


stp_product_definition_shape * find_owning_pdef_shape ( 
    stp_representation * rep
    )
{
    unsigned i,sz;
    RoseBackptrs *bp = rose_get_backptrs(rep);

    // Look through all properties that this representation is used by
    for (i=0, sz=bp->size(); i<sz; i++) {
	RoseObject * ref = bp->get(i);
	if (ref-> isa(ROSE_DOMAIN(stp_property_definition_representation))) {
	    stp_property_definition_representation * pdr = 
		ROSE_CAST(stp_property_definition_representation,ref);

	    if (pdr->definition() &&
		pdr->definition()-> is_property_definition()) {
		stp_property_definition * pd = 
		    pdr->definition()-> _property_definition();

		if (pd-> isa(ROSE_DOMAIN(stp_product_definition_shape)))
		    return ROSE_CAST(stp_product_definition_shape,pd);
	    }
	}
    }
    return 0;
}



//============================================================
// FIND REPRESENTATION CONTEXT -- Find the representation context used
// for a given representation item.  If it appears in multiple reps,
// return the first one.
//
stp_representation_context * find_rep_context_for_repitem (
    stp_representation_item * it
    )
{
    rose_vector reps;

    ROSE.beginTraversal();
    find_owning_representations (it, reps);
    ROSE.endTraversal();

    stp_representation * rep = (stp_representation*) reps[0];

    return (rep? rep-> context_of_items(): 0);
}
